# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django.http import Http404
from django.views.decorators.csrf import csrf_exempt
# from jose.exceptions import (
#     ExpiredSignatureError,
#     JWTClaimsError,
#     JWTError,
# )
from django.contrib.auth import authenticate
from decimal import Decimal
from query_designer.api import *
from query_designer.lists import AGGREGATES
import json
from django.db import transaction
from django.http import HttpResponse, JsonResponse, HttpResponseRedirect
from django.shortcuts import redirect, render
from django.conf import settings
from datetime import datetime, timedelta
import time
from django.contrib.auth.decorators import login_required
from django.core.exceptions import PermissionDenied
from django.views.decorators.cache import never_cache
from django.core.management import call_command
from django.urls.base import reverse
from django.contrib.auth import logout
import requests

# @login_required
@never_cache
def load_query(request, pk):
    q = Query.objects.get(pk=pk)

    return JsonResponse({
        'pk': q.pk,
        'title': q.title,
        'design': q.design,
        'document': q.document
    })


# @login_required
@never_cache
def clean(request, pk=None):
    user, token = get_user(request)
    storage_target = settings.STORAGE_TARGET

    dataset_list = Dataset.objects.filter(hidden=False)

    # dataset_list.order_by('order')
    publisher_list = sorted(
        set([d.publisher for d in dataset_list if d.publisher is not None and len(str(d.publisher).strip()) > 0]))
    organization_list = sorted(
        set([d.publisher for d in dataset_list if d.publisher is not None and len(str(d.publisher).strip()) > 0]))
    observation_list = sorted(
        set([d.observations for d in dataset_list if d.observations is not None]))
    license_list = sorted(
        set([d.license for d in dataset_list if d.license is not None]))
    stored_at_list = sorted(
        set([d.stored_at for d in dataset_list if d.stored_at is not None]))

    variable_list = Variable.objects.filter(
        dataset__hidden=False).order_by('title')
    variable_list_filter = sorted(
        set([(v.name, v.title) for v in variable_list]))

    try:
        time_start_timestamp = min([d.temporalCoverageBeginTimestamp for d in Dataset.objects.all(
        ) if d.temporalCoverageBeginTimestamp != ""])
    except:
        # TO BE CHANGED!!
        time_start_timestamp = int(datetime.timestamp(
            datetime.now() - timedelta(days=1000))) * 1000

    date_now = datetime.now()
    time_end_timestamp = max([d.temporalCoverageEndTimestamp for d in Dataset.objects.all(
    ) if d.temporalCoverageEndTimestamp != ""] + [int(time.mktime(date_now.timetuple())) * 1000])

    # dataset_list = dataset_list.order_by('order', 'title')

    return render(request, 'query_designer/clean.html', {
        'organizations': organization_list,
        'publishers': publisher_list,
        'observations': observation_list,
        'licenses': license_list,
        'stored_at_list': stored_at_list,
        'variables': variable_list,
        'variable_list_filter': variable_list_filter,
        'datasets': dataset_list,
        'dimensions': Dimension.objects.all(),
        'available_viz': [],
        'AGGREGATES': AGGREGATES,
        'time_start_timestamp': time_start_timestamp,
        'time_end_timestamp': time_end_timestamp,
        'settings': {'PLATFORM_VERSION': settings.PLATFORM_VERSION}
    })


# @login_required
@csrf_exempt
def save_query(request, pk=None, temp=1):
    # print pk, temp
    # create or update
    if not pk:
        user, token = get_user(request)
        if user == 'redirect':
            return JsonResponse({
                'redirect': 'true',
                'url': reverse('clean')
            })
        if user.is_authenticated:
            if int(temp) == 1:
                # print 'it is temp'
                q = TempQuery(user=user)
            else:
                q = Query(user=user)
        else:
            if int(temp) == 1:
                q = TempQuery()
            else:
                q = Query()
    else:
        # print 'not pk'
        q = AbstractQuery.objects.get(pk=pk)

    # document
    if 'document' in request.POST:
        try:
            doc = json.loads(request.POST.get('document', ''))
        except ValueError:
            return JsonResponse({'error': 'Invalid query document'}, status=400)

        q.document = doc

    # title
    if 'title' in request.POST:
        q.title = request.POST.get('title', 'Untitled query')

    # save
    q.save()

    # return OK response
    return JsonResponse({
        'pk': q.pk,
        'title': q.title,
    })


@login_required
@never_cache
def get_query_variables(request):
    if 'id' in request.GET:
        id = int(request.GET.get('id'))
        doc = None
        # existing saved query
        if id >= 0:
            # get the query
            q = AbstractQuery.objects.get(pk=id)
            # get the document
            doc = q.document
        # new query
        else:
            if 'document' in request.GET:
                try:
                    doc = json.loads(request.GET.get('document', ''))
                except ValueError:
                    return JsonResponse({'error': 'Invalid query document'}, status=400)

        # find the query's SELECT variables formatted as: 'variable's column name (title)': 'variable's name'
        variables = dict()
        dimensions = dict()
        for from_clause in doc['from']:
            for col in from_clause['select']:
                if col['type'] == 'VALUE':
                    variables[col['title']] = col['name']
                else:
                    dimensions[col['title']] = col['name']

        result = {'variables': variables, 'dimensions': dimensions}
        # return the found variables
        return HttpResponse(json.dumps(result), content_type="application/json")

    else:
        return JsonResponse({'error': 'Invalid query document'}, status=400)


class DefaultEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, Decimal):
            return float(o)
        elif isinstance(o, datetime.datetime):
            return o.isoformat()
        elif isinstance(o, datetime.date):
            return o.isoformat()

        return super(DefaultEncoder, self).default(o)


def filter_info(request, filter_type, pk):
    if filter_type == 'variable':
        return JsonResponse({'type': 'text'})
        # return JsonResponse({'type': 'number', 'orderable': True})

    dimension = Dimension.objects.get(pk=pk)

    # values = dimension.ranges
    # TODO: change this, get proper dimension ranges
    values = {
        'type': 'text'
    }
    return JsonResponse(values, encoder=DefaultEncoder)

    if type(values) == dict and 'min' in values:
        values.update({
            'type': 'text',
            'orderable': True,
        })

        return JsonResponse(values, encoder=DefaultEncoder)

    return JsonResponse({
        'type': 'select',
        'orderable': True,
        'options': [{
            'name': v,
            'value': v,
        } for v in values]
    }, encoder=DefaultEncoder)


def get_field_policy():
    field_policy = {
        'categories': [],
        'variables': [],
        'aggregates': [],
    }

    # for variable in Variable.objects.all():
    #     field_policy['variables'].append({
    #         'name': variable.name,
    #         'title': variable.title,
    #         'id': variable.pk,
    #         'description': variable.description,
    #         'dataType': variable.dataType,
    #     })

    for aggregate in AGGREGATES:
        field_policy['aggregates'].append({
            'title': aggregate[1],
            'value': aggregate[0],
            'typelist': aggregate[2],
        })

    return field_policy


def get_config(request):
    return JsonResponse(get_field_policy())


@never_cache
@login_required
def list_queries(request):
    user, token = get_user(request)
    if user == 'redirect':
        return JsonResponse({
            'redirect': 'true',
            'url': reverse('clean')
        })

    # ensure GET request
    if request.method != 'GET':
        return HttpResponse('Only `GET` method allowed', status=400)

    ctx = {
        'queries': Query.objects.filter(user=user).order_by().order_by('-created', '-updated'),
    }

    return render(request, 'query_designer/utils/query-table.html', ctx)


# @login_required
def delete_query(request, pk):
    user, token = get_user(request)
    if user == 'redirect':
        return JsonResponse({
            'redirect': 'true',
            'url': reverse('clean')
        })
    # ensure DELETE request
    if request.method != 'DELETE':
        return HttpResponse('Only `DELETE` method allowed', status=400)

    if user is not None:
        try:
            query = AbstractQuery.objects.get(pk=int(pk), user=user)
            query.delete()
            return HttpResponse('Query deleted', status=200)
        except AbstractQuery.DoesNotExist as e:
            return HttpResponse('Query not found', status=404)
    else:
        return HttpResponse('Uknown user', status=400)

@never_cache
# @login_required
def api_list_user_queries(request):
    user, token = get_user(request)
    if user == 'redirect':
        return JsonResponse({
            'redirect': 'true',
            'url': reverse('clean')
        })

    # if user is not None:
    queries = Query.objects.filter(user=user).values(
        'id', 'title', 'created', 'updated').order_by().order_by('-created', '-updated')
    # json_queries = json.dumps(list(queries))
    return JsonResponse(list(queries), safe=False)
    # else:
    #     queries = Query.objects.filter(generated_by='CUSTOM').values(
    #         'id', 'title', 'created', 'updated').order_by().order_by('-created', '-updated')
    #     return JsonResponse(list(queries), safe=False)
    #     # return HttpResponse('Uknown user', status=400)


@never_cache
# @login_required
def get_query_statement(request, query_id):
    user, token = get_user(request)
    if user == 'redirect':
        return JsonResponse({
            'redirect': 'true',
            'url': reverse('clean')
        })
    # print(user)

    # ensure GET request
    if request.method != 'GET':
        return HttpResponse('Only `GET` method allowed', status=400)
    try:
        query = Query.objects.get(pk=int(query_id))
        if query.user != user:
            raise PermissionDenied
        return JsonResponse({'query_statement': query.raw_query}, safe=False)
    except Query.DoesNotExist as e:
        raise Http404


@never_cache
# @login_required
def get_query_info(request, query_id):
    user, token = get_user(request)
    if user == 'redirect':
        return JsonResponse({
            'redirect': 'true',
            'url': reverse('clean')
        })

    # ensure GET request
    if request.method != 'GET':
        return HttpResponse('Only `GET` method allowed', status=400)

    # if user is not None:
    try:
        query = Query.objects.get(pk=int(query_id))
        # if query.user != user:
        #     raise PermissionDenied

        variables = list()
        dimensions = list()
        doc = query.document
        # print (doc)
        for from_clause in doc['from']:
            for col in from_clause['select']:
                if not col['exclude']:
                    if col['type'] == 'VALUE':
                        variables.append({'title': col['title'], 'name_in_query': col['name'], 'column_in_db': Variable.objects.get(
                            pk=int(from_clause['type'])).name})
                    else:
                        dimensions.append({'title': col['title'], 'name_in_query': col['name'], 'column_in_db': Dimension.objects.get(
                            pk=int(col['type'])).name})

        columns = {'variables': variables, 'dimensions': dimensions}

        return JsonResponse({
            'id': query.id,
            'title': query.title,
            'created': query.created,
            'updated': query.updated,
            'columns': columns,
            'query_statement': query.raw_query,
            'document': query.document,
        }, safe=False)
    except Query.DoesNotExist as e:
        raise Http404
    # else:
    #     return HttpResponse('Uknown user', status=400)


def get_config(request):
    return JsonResponse(get_field_policy())

